#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import deque
import itertools
import textwrap
import tomllib
import unittest

from balladeer.examples.ex_10_lite_sequence.app import Story as Story_10
from balladeer.lite.loader import Loader
from balladeer.lite.speech import Dialogue
from balladeer.lite.story import StoryBuilder
from balladeer.lite.types import Grouping


class StoryTests(unittest.TestCase):
    def test_simple_turns(self):
        story = StoryBuilder(
            Dialogue("<> Knock, knock."),
            Dialogue("<> Who's there?"),
        )
        self.assertIsInstance(story.context.speech, deque)
        for n in range(3):
            with (
                self.subTest(n=n),
                story.turn() as turn
            ):
                if n < 2:
                    self.assertTrue(turn.blocks, turn)
                else:
                    self.assertFalse(turn.blocks, turn)

class ExampleTests(unittest.TestCase):
    def test_cartoon_fight(self):
        content = textwrap.dedent('''
            [FIGHTER_1]
            state = 1

            [FIGHTER_2]
            type = "Animal"
            state = 1

            [WEAPON]
            type = "Weapon"
            roles = ["FIGHTER_1"]
            state = 1

            [[_]]

            s="""
            <FIGHTER_1>

                I don't like the way you use me, {FIGHTER_2.name}!

            <WEAPON.attacking@FIGHTER_2:shouts/slapwhack?offer=1>

                _Whack!_

            <FIGHTER_2?offer=1>

                Uuurrggh!
            """

            ''').strip()

        scene = Loader.read(content)
        story = Story_10(config={}, assets=Grouping.typewise([scene]))
        specs = story.director.specifications(scene.tables)
        self.assertEqual(story.director.rank_constraints(specs["FIGHTER_1"]), 1, specs)
        self.assertEqual(story.director.rank_constraints(specs["FIGHTER_2"]), 2, specs)
        self.assertEqual(story.director.rank_constraints(specs["WEAPON"]), 3, specs)

        for n in range(3):
            with self.subTest(n=n, m=len(story.director.notes[(None, 0)].maps)):
                notes = story.director.notes[(None, 0)]

                with story.turn() as turn:

                    roles = dict(story.director.roles(specs, story.context.ensemble))
                    if not n:
                        self.assertTrue(turn.blocks, notes)
                        self.assertEqual(3, len(roles), roles)
                    elif n == 1:
                        self.assertTrue(turn.blocks, notes)
                        self.assertEqual(2, len(roles), roles)
                        rewriter = story.director.rewrite(scene, roles)
                        self.assertRaises(KeyError, list, rewriter)
                    elif n == 2:
                        self.assertFalse(turn.blocks, notes)




#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from collections.abc import Generator
from collections import ChainMap
from collections import Counter
from collections import defaultdict
import html
import itertools
import pathlib
import re
import string
import urllib.parse


from balladeer.lite.entity import Entity
from balladeer.lite.loader import Loader
from balladeer.lite.speech import Speech
from balladeer.lite.speech import Prologue
from balladeer.lite.speech import Dialogue
from balladeer.lite.speech import Epilogue
from balladeer.lite.types import Grouping


class Director:
    class Formatter(string.Formatter):
        def convert_field(self, value: str, conversion: str) -> str:
            if conversion != "a":
                return super().convert_field(value, conversion)
            else:
                return value.translate(Speech.processor.escape_table)

    @staticmethod
    def specify_role(spec: dict) -> tuple[set, set, dict]:
        roles = set(filter(None, spec.get("roles", []) + [spec.get("role")]))
        types = set(filter(None, spec.get("types", []) + [spec.get("type")]))
        states = {k: set(v) for k, v in spec.get("states", {}).items()}
        try:
            key, value = spec["state"].split(".")
            states.setdefault(key, set()).add(value)
        except AttributeError:
            # Special case for integer state
            value = spec["state"]
            values = set([value]) if isinstance(value, int) else set(value)
            states["int"] = states.get("int", set()).union(values)
        except (KeyError, ValueError):
            pass

        return roles, states, types

    def __init__(
        self,
        shot_key: str = "_",
        dialogue_key: str = "s",
        pause: float = 1,
        dwell: float = 0.1,
        delay: float = 0,
        offer: float = None,
    ):
        self.fmtr = self.Formatter()
        self.counts = Counter()
        self.notes = defaultdict(ChainMap)

        self.shot_key = shot_key
        self.dialogue_key = dialogue_key
        self.pause = pause
        self.dwell = dwell
        self.delay = delay
        self.offer = offer

        self.cast = None
        self.role = None

        self.bq_matcher = re.compile("<blockquote.*?<\\/blockquote>", re.DOTALL)
        self.tag_matcher = re.compile("<[^>]+?>")
        self.cite_matcher = re.compile(
            """
            (?P<head><cite.*?data-role=")   # Up until role attribute
            (?P<role>[^"]+?)                # Role attribute
            (?P<tail>"[^>]*?)               # Until end of opening tag
            >                               # Close opening tag
            .*?</cite>                      # Text and closing tag
            """,
            re.VERBOSE,
        )
        self.attr_matcher = re.compile('data-([^=]+)=[^"]*"([^"]*)"')
        self.ul_matcher = re.compile("<ul>.*?<\\/ul>", re.DOTALL)
        self.li_matcher = re.compile("<li>.*?<\\/li>", re.DOTALL)
        self.pp_matcher = re.compile("<p>(.*?)<\\/p>", re.DOTALL)

    def attributes(self, text: str) -> dict[str, str]:
        return dict(self.attr_matcher.findall(text))

    def fragments(self, attrs: dict) -> str:
        # NOTE: Sunject to change
        return html.unescape(attrs.get("fragments", "")).lstrip("#")

    def directives(self, attrs: dict) -> dict[str, list[str]]:
        lhs, _, rhs = html.unescape(attrs.get("directives", "")).partition("@")
        return {d: [i for i in rhs.split(",") if i] for d in lhs.split(".") if d}

    def mode(self, attrs: dict) -> tuple[str]:
        return tuple(
            i for i in html.unescape(attrs.get("mode", "")).lstrip(":").split("/") if i
        )

    def parameters(self, attrs: dict) -> dict:
        params = urllib.parse.parse_qs(html.unescape(attrs.get("parameters", "").lstrip("?")))
        try:
            offer = float(params.get("offer", [self.offer])[0])
            if offer < 0:
                offer = None
        except TypeError:
            offer = None

        return {
            "pause": float(params.get("pause", [self.pause])[0]),
            "dwell": float(params.get("dwell", [self.dwell])[0]),
            "delay": float(params.get("delay", [self.delay])[0]),
            "offer": offer,
        }

    def handle_fragments(self, html5, fragments: str, path: pathlib.Path | str, index: int):
        if not fragments:
            return html5
        elif fragments.isdigit():
            self.notes[(path, index)]["option"] = int(fragments)
        elif fragments.endswith("!"):
            list_block = self.ul_matcher.search(html5)
            list_items = list(self.li_matcher.findall(list_block.group()))
            ordinal = self.counts[(path, index)] % len(list_items)
            choice = list_items[ordinal].replace("<li>", "").replace("</li>", "")
            html5 = self.ul_matcher.sub(choice, html5)
            self.counts[(path, index)] += 1

        return html5

    def handle_directives(self, html5, directives: dict, path: pathlib.Path | str, index: int):
        self.notes[(path, index)]["directives"] = [
            (
                directive,
                self.cast.get(self.role),
                tuple(filter(None, (self.cast.get(r) for r in roles))),
            )
            for directive, roles in directives.items()
        ]
        return html5

    def handle_mode(self, html5, mode: list[str], path: pathlib.Path | str, index: int):
        if mode:
            self.notes[(path, index)]["mode"] = mode[0]
            self.notes[(path, index)]["media"] = [i for n, i in enumerate(mode) if n]

        return html5

    def handle_parameters(self, html5, parameters: dict, path: pathlib.Path | str, index: int):
        self.pause = parameters["pause"]
        self.dwell = parameters["dwell"]
        self.offer = parameters["offer"]
        self.notes[(path, index)]["offer"] = self.offer
        return html5

    def rank_constraints(self, spec: dict) -> int:
        roles, states, types = self.specify_role(spec)
        return sum(1 / len(v) for v in states.values()) + len(types) + len(roles)

    def edit(
        self,
        speech: Speech,
        roles: dict = {},
        path: pathlib.Path | str = None,
        index: int = 0,
    ) -> Generator[str]:
        self.cast = roles.copy()
        for block in self.bq_matcher.findall(speech.tags):
            self.notes[(path, index)] = self.notes[(path, index)].new_child(
                pause=self.pause, duration=self.dwell, delay=self.delay
            )

            html5 = self.cite_matcher.sub(self.edit_cite, block)

            attrs = self.attributes(html5)

            parameters = self.parameters(attrs)
            html5 = self.handle_parameters(html5, parameters, path, index)

            fragments = self.fragments(attrs)
            html5 = self.handle_fragments(html5, fragments, path, index)

            mode = self.mode(attrs)
            html5 = self.handle_mode(html5, mode, path, index)

            directives = self.directives(attrs)
            html5 = self.handle_directives(html5, directives, path, index)

            html5 = self.pp_matcher.sub(self.edit_para, html5)

            yield self.fmtr.format(html5, **self.cast)

    def edit_cite(self, match: re.Match) -> str:
        key = list(self.notes)[-1]
        self.notes[key]["type"] = "cue"
        head, self.role, tail = [match.group(i) for i in ("head", "role", "tail")]
        try:
            entity = self.cast[self.role]
        except KeyError:
            return match.group()

        #TODO: New child for notes.
        delay = self.delay + self.pause
        try:
            attr = f'" data-entity="{entity.names[0]}'
            text = entity.names[0].translate(Speech.processor.escape_table)
            self.notes[key]["entity"] = entity
            self.notes[key]["role"] = self.role
            return (
                f'{head}{self.role}{attr}{tail}'
                f' style="animation-delay: {delay:.2f}s; animation-duration: {self.dwell:.2f}s">'
                f'{text}</cite>'
            )
        except IndexError:
            return match.group()

    def edit_para(self, match: re.Match) -> str:
        content = match.group(1).strip()
        if not content:
            return ""

        words = Speech(content).words
        delay = self.delay + self.pause
        duration = self.dwell * len(words)
        self.delay = delay + duration

        key = list(self.notes)[-1]
        self.notes[key] = self.notes[key].new_child(pause=self.pause, duration=duration, delay=self.delay)
        return (
            f'<p style="animation-delay: {delay:.2f}s; animation-duration:'
            f' {duration:.2f}s">{content}</p>'
        )

    def roles(self, specs: dict, ensemble: list[Entity]) -> dict[str, Entity]:
        specs = dict(
            sorted(specs.items(), key=lambda x: self.rank_constraints(x[1]), reverse=True)
        )
        pool = {i: set(specs.keys()) for i in ensemble}
        for role, spec in specs.items():
            roles, states, types = self.specify_role(spec)
            try:
                entity = next(
                    entity
                    for entity, jobs in pool.items()
                    if role in jobs
                    and (
                        not types
                        or entity.types.intersection(types)
                        or entity.__class__.__name__ in types
                    )
                    and all(
                        k in entity.states
                        and getattr(entity.get_state(k), "name", entity.get_state(k)) in v
                        for k, v in states.items()
                    )
                )
            except StopIteration:
                continue
            else:
                pool[entity] = roles
                yield role, entity

    def specifications(self, toml: dict):
        return {k: v for k, v in toml.items() if isinstance(v, dict) and k != self.shot_key}

    def selection(self, scripts, ensemble: list[Entity] = [], roles=1):
        for scene in scripts:
            specs = self.specifications(scene.tables)
            roles = dict(self.roles(specs, ensemble))
            if len(roles) == len(specs):
                return scene, specs, roles
        else:
            return Loader.Scene("", {}), {}, {}

    def allows(self, conditions: dict, cast: dict[str, Entity] = {}) -> bool:
        for role, (roles, states, types) in conditions.items():
            entity = cast[role]
            if (
                types
                and not types.issubset(entity.types)
                and entity.__class__.__name__ not in types
            ):
                return False
            for state, values in states.items():
                try:
                    key = entity.get_state(state).name
                except AttributeError:
                    key = entity.get_state(state)

                if key not in values:
                    return False

        return True

    def specify_conditions(self, shot: dict) -> Generator[tuple[str, tuple[set, set, dict]]]:
        for role, guard in shot.get("if", {}).items():
            yield role, self.specify_role(guard)

    def dialogue(self, scene: Loader.Scene, roles: dict) -> tuple[int, Dialogue]:
        shots = scene.tables.get(self.shot_key, [])
        for n, shot in enumerate(shots):
            conditions = dict(self.specify_conditions(shot))
            if self.allows(conditions, roles):
                text = shot.get(self.dialogue_key, "")
                yield n, Dialogue(text)

    def rewrite(
            self,
            scene: Loader.Scene = Loader.Scene("", {}),
            roles: dict[str, Entity] = {},
            speech: list[Speech] = [],
        ) -> Generator[tuple[int | None, str]]:

        n_shots = len(scene.tables.get(self.shot_key, [])) if scene else 0
        shot_tuples = [
            (t, n_shots + n, s)
            for n, (t, s) in enumerate(
                (t, s)
                for t, v in Grouping.typewise(speech).items()
                for s in v
            )
        ]
        spoken = Grouping(list)
        for t, n, s in shot_tuples:
            for b in self.edit(s, roles, path=None, index=n):
                spoken[t].append((n, b))

        spoken[Dialogue] = list(itertools.chain.from_iterable(itertools.zip_longest(
            spoken.setdefault(Dialogue, []),
            [
                (n, html5)
                for n, d in self.dialogue(scene, roles)
                for html5 in self.edit(d, roles, path=scene.path, index=n)
            ]
        )))

        self.delay = 0

        f = lambda x: x and x[1]
        yield from filter(f, spoken.get(Prologue, []))
        yield from filter(f, spoken.get(Dialogue, []))
        yield from filter(f, spoken.get(Epilogue, []))



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import unittest

from balladeer.lite.speech import Speech


class SpeechTests(unittest.TestCase):

    def test_trim(self):
        text = """
            1
            2
            3
        """
        s = Speech(text)
        a = s.trim()
        b = s.trim()
        self.assertIs(a, b)
        self.assertEqual(5, len(a), a)

    def test_tags(self):
        text = """
            <> T

            <>
            1. a
            2. b
            3. c

        """
        s = Speech(text)
        a = s.tags
        b = s.tags
        self.assertIs(a, b)

    def test_line_count(self):
        text = """
        <STAFF.proposing#3> What will you have, sir? The special is fish today.

            1. Order the Beef Wellington
            2. Go for the Shepherd's Pie
            3. Try the Dover Sole
        """
        s = Speech(text)

        a = s.lines
        b = s.lines
        self.assertIs(a, b)
        self.assertEqual(5, len(s.lines))

    def test_word_count(self):
        text = """
        <STAFF.proposing#3> What will you have, sir? The special is fish today.

            1. Order the Beef Wellington
            2. Go for the Shepherd's Pie
            3. Try the Dover Sole
        """
        s = Speech(text)

        a = s.words
        b = s.words
        self.assertIs(a, b)
        self.assertEqual(24, len(s.words))




#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum
import json
import unittest

from balladeer.lite.entity import Entity
from balladeer.lite.types import State


class TestComparisons(unittest.TestCase):
    def test_identity(self):
        bag = set(
            [
                Entity(),
                Entity(),
                Entity(),
            ]
        )
        self.assertEqual(3, len(bag))

    def test_quality(self):
        a = Entity(name="a")
        b = Entity(name="a")
        self.assertEqual(a, b)

    def test_inequality(self):
        a = Entity(name="a")
        b = Entity(name="b")
        self.assertNotEqual(a, b)


class TestAttributes(unittest.TestCase):
    def test_aspect(self):
        a = Entity(aspect="whistling")
        b = Entity(aspect="whistling")
        self.assertNotEqual(a, b)

    def test_sketch(self):
        a = Entity(sketch="athletic")
        b = Entity(sketch="athletic")
        self.assertNotEqual(a, b)


class TestNamesAndTypes(unittest.TestCase):
    class Active(State):
        off = 0
        on =  1

    class Thing(Entity):
        pass

    def test_string_type(self):
        e = Entity(type="Thing")
        self.assertEqual({"Thing"}, e.types)

    def test_declared_type(self):
        e = Entity(type=TestNamesAndTypes.Thing)
        self.assertEqual({"Thing"}, e.types)

    def test_dumps_to_json(self):
        e = Entity(type=TestNamesAndTypes.Thing).set_state(3, TestNamesAndTypes.Active.on)
        self.assertEqual({"Thing"}, e.types)

        j = Entity.Encoder().encode(e)
        self.assertTrue(j)

        s = json.loads(j)
        self.assertTrue(s)


class TestEnumStates(unittest.TestCase):
    class Colour(enum.Enum):
        red = "ff0000"
        green = "00ff00"
        blue = "0000ff"

    def test_state_as_enum(self):
        s = Entity()
        s.set_state(TestEnumStates.Colour.red)
        self.assertEqual(TestEnumStates.Colour.red, s.get_state(TestEnumStates.Colour))
        self.assertEqual({"Colour": TestEnumStates.Colour.red}, s.states)
        j = Entity.Encoder().encode(s)
        self.assertTrue(j)

    def test_state_as_enum_twice(self):
        s = Entity()
        s.set_state(
            TestEnumStates.Colour.red,
            TestEnumStates.Colour.red,
        )
        self.assertEqual(TestEnumStates.Colour.red, s.get_state(TestEnumStates.Colour))
        self.assertEqual({"Colour": TestEnumStates.Colour.red}, s.states)

    def test_state_as_enum_args(self):
        s = Entity()
        s.set_state(
            TestEnumStates.Colour.red,
            TestEnumStates.Colour.green,
        )
        self.assertEqual(TestEnumStates.Colour.green, s.get_state(TestEnumStates.Colour))
        self.assertEqual({"Colour": TestEnumStates.Colour.green}, s.states)


class TestIntegerStates(unittest.TestCase):
    def test_state_as_int(self):
        s = Entity()
        s.set_state(3)
        self.assertEqual(3, s.get_state(int))
        self.assertEqual(3, s.get_state("int"))
        self.assertEqual(3, s.state)

    def test_state_as_int_twice(self):
        s = Entity()
        s.set_state(3, 3)
        self.assertEqual(3, s.get_state(int))
        self.assertEqual(3, s.get_state("int"))
        self.assertEqual(3, s.state)

    def test_state_as_int_args(self):
        s = Entity()
        s.set_state(3, 4, 5)
        self.assertEqual(5, s.get_state(int))
        self.assertEqual(5, s.get_state("int"))
        self.assertEqual(5, s.state)


class TestStringStates(unittest.TestCase):
    class Colour(enum.Enum):
        red = "ff0000"
        green = "00ff00"
        blue = "0000ff"

    def test_get_state_as_string(self):
        e = Entity()
        e.set_state(TestEnumStates.Colour.green)

        self.assertEqual(TestEnumStates.Colour.green, e.get_state("Colour"))
        self.assertIsNone(e.get_state("Color"))



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import unittest

from balladeer.lite.entity import Entity
from balladeer.lite.loader import Loader
from balladeer.lite.types import Grouping


class GroupingTests(unittest.TestCase):

    def test_grouping_by_class(self):
        class A: pass
        class B: pass

        g = Grouping.typewise([A(), B(), A()])
        self.assertIn(A, g)
        self.assertEqual(2, len(g[A]))

        self.assertIn(B, g)
        self.assertEqual(1, len(g[B]))
        self.assertEqual(3, len(g.all))
        self.assertEqual(3, len(g.each))
 
    def test_grouping_by_type(self):
        types = ("audio/mpeg", "text/css", "text/javascript")
        items = [
            Loader.Asset(unittest, "data/slapwhack.mp4", type=types[0]),
            Loader.Asset(unittest, "styles/style.css", type=types[1]),
            Loader.Asset(unittest, "picker.js", type=types[2]),
        ]
        g = Grouping.typewise(items)

        for t in types:
            with self.subTest(t=t):
                self.assertIn(t, g)
                self.assertEqual(1, len(g[t]))
                self.assertIsInstance(g[t][0], Loader.Asset)

    def test_grouping_via_types(self):
        items = [
            Entity(type="A"),
            Entity(type="B"),
            Entity(type="A"),
        ]

        g = Grouping.typewise(items)

        self.assertIn(Entity, g)
        self.assertEqual(3, len(g[Entity]))

        self.assertIn("A", g)
        self.assertEqual(2, len(g["A"]))

        self.assertIn("B", g)
        self.assertEqual(1, len(g["B"]))

        self.assertEqual(6, len(g.all))
        self.assertEqual(3, len(g.each))

    def test_grouping_by_multiple_type(self):
        items = [
            Entity(type="A"),
            Entity(types={"A", "B"}),
        ]

        g = Grouping.typewise(items)

        self.assertIn(Entity, g)
        self.assertEqual(2, len(g[Entity]))

        self.assertIn("A", g)
        self.assertEqual(2, len(g["A"]))

        self.assertIn("B", g)
        self.assertEqual(1, len(g["B"]))

        self.assertEqual(5, len(g.all), g.all)
        self.assertEqual(2, len(g.each))



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import defaultdict
import enum

from balladeer.lite.entity import Entity
from balladeer.lite.types import State


Into = None
Spot = None
Exit = None


class Traffic(State, enum.Enum):
    blocked = enum.auto()
    forward = enum.auto()
    reverse = enum.auto()
    flowing = enum.auto()


class Transit(Entity):
    pass


# TODO: Reconcile with balladeer.cartography.Map
class MapMaker:
    def __init__(self, spots, config=None):
        self.config = config
        global Into, Spot, Exit
        self.into = Into = enum.Enum("Into", spots, type=State)
        self.exit = Exit = enum.Enum("Exit", spots, type=State)
        self.spot = Spot = enum.Enum("Spot", spots, type=State)
        self.transits = list(self.make())

    def make(self):
        raise NotImplementedError


class WorldBuilder:

    def __init__(self, map, config=None):
        self.map = map
        self.config = config
        # TODO: Grouper by type or name of type?
        self.entities = list(self.build())

    def build(self):
        return ()



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import dataclasses
import enum
import json
import random
import re
import uuid

from balladeer.lite.speech import Speech
from balladeer.lite.types import State


@dataclasses.dataclass(kw_only=True, unsafe_hash=True)
class Entity:
    name: dataclasses.InitVar = ""
    names: list = dataclasses.field(default_factory=list, compare=False)

    type: dataclasses.InitVar = ""
    types: set = dataclasses.field(default_factory=set, compare=False)

    states: dict = dataclasses.field(default_factory=dict, compare=False)
    uid: uuid.UUID = dataclasses.field(default_factory=uuid.uuid4)

    sketch: Speech = dataclasses.field(default_factory=Speech, compare=False)
    aspect: Speech = dataclasses.field(default_factory=Speech, compare=False)

    def __post_init__(self, name, type):
        # Unfortunately the builtin 'type' can't be used here
        if name and not isinstance(name, property):
            self.names.insert(0, name)
        if type and not isinstance(type, property):
            try:
                self.types.add(type.__name__)
            except AttributeError:
                self.types.add(type)

    def __eq__(self, other):
        if not self.names:
            return self.uid == other.uid

        try:
            return set(self.names).union(self.types) == set(other.names).union(other.types)
        except AttributeError:
            return False

    class Encoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, enum.Enum):
                return {"name": obj.name, "value": obj.value}

            data = dataclasses.asdict(obj)
            data["uid"] = str(data["uid"])
            data["types"] = sorted(data.get("types", []))
            return data

    @property
    def name(self):
        return random.choice(self.names or [""])

    @property
    def type(self):
        return sorted(self.types)[0] if self.types else type(self)

    @property
    def state(self):
        return self.get_state()

    @state.setter
    def state(self, value):
        return self.set_state(value)

    def set_state(self, *args):
        for arg in args:
            self.states[type(arg).__name__] = arg
        return self

    def get_state(self, typ: State = int):
        try:
            return self.states.get(typ.__name__)
        except AttributeError:
            return self.states.get(typ)
        else:
            return None

    @property
    def state(self):
        return self.get_state()

    @state.setter
    def state(self, value):
        return self.set_state(value)

    def set_state(self, *args):
        for arg in args:
            self.states[type(arg).__name__] = arg
        return self

    def get_state(self, typ: State = int):
        try:
            return self.states.get(typ.__name__)
        except AttributeError:
            return self.states.get(typ)
        else:
            return None



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import deque
from collections import namedtuple
from collections.abc import Callable
import uuid
import warnings

from balladeer.lite.director import Director
from balladeer.lite.drama import Drama
from balladeer.lite.loader import Loader
from balladeer.lite.speech import Speech
from balladeer.lite.types import Grouping
from balladeer.lite.world import WorldBuilder


class StoryBuilder:

    Turn = namedtuple(
        "Turn",
        ["scene", "specs", "roles", "speech", "blocks", "notes"],
        defaults=(None, None)
    )

    def __init__(
        self,
        *speech: tuple[Speech],
        config = None,
        assets: Grouping = Grouping(),
        world: WorldBuilder = None, drama: [list | deque] = None,
        **kwargs
    ):
        self.uid = uuid.uuid4()
        self.speech = speech
        self.config = config
        self.assets = assets.copy()
        if not world:
            world_type = next(reversed(WorldBuilder.__subclasses__()), WorldBuilder)
            self.world = world_type(config)
        else:
            self.world = world

        self.drama = drama or deque([])
        self.drama.extend(self.build())

        self.director = Director(**kwargs)

    def build(self):
        drama_classes = Drama.__subclasses__()
        if self.speech or not drama_classes:
            yield Drama(
                *self.speech,
                world=self.world,
                config=self.config
            )
        else:
            yield from (d(world=self.world, config=self.config) for d in drama_classes)

    @property
    def context(self):
        return self.drama[0]

    @property
    def notes(self):
        return list(self.director.notes.values())

    @property
    def direction(self):
        notes = self.notes
        if not notes:
            return []
        else:
            return [t for i in (m.get("directives", []) for m in notes[-1].maps) for t in i]

    def action(self, text: str, *args, **kwargs):
        drama = self.context
        actions = drama.actions(
            text,
            context=self.director,
            ensemble=drama.ensemble,
            prefix=drama.prefixes[0],
        )
        fn, args, kwargs = drama.pick(actions)
        if not fn:
            return None

        try:
            drama.speech.extend(drama(fn, *args, **kwargs))
        except Exception as e:
            warnings.warn(e)

        return fn, args, kwargs

    def turn(self, *args, **kwargs):
        self.context.interlude(*args, **kwargs)
        return self

    def __enter__(self):
        drama = self.context

        # Director selection
        scripts = drama.scripts(self.assets.get(Loader.Scene, []))
        scene, specs, roles = self.director.selection(scripts, drama.ensemble)
        assert isinstance(scene, Loader.Scene), f"{type(scene)} is not a Scene"

        # TODO: Entity aspects

        # Collect waiting speech
        try:
            speech = [drama.speech.popleft()]
        except AttributeError:
            speech = drama.speech.copy()
            drama.speech.clear()
        except IndexError:
            speech = []

        blocks = list(self.director.rewrite(scene, roles, speech))

        # Directive handlers
        for action, entity, entities in self.direction:
            method = getattr(drama, f"{drama.prefixes[1]}{action}")
            if isinstance(method, Callable):
                try:
                    method(entity, *entities)
                except Exception as e:
                    warnings.warn(e)

        return self.Turn(scene, specs, roles, speech, blocks, self.director.notes.copy())

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.director.notes.clear()
        return False



from turberfield.catchphrase.parser import CommandParser

from turberfield.catchphrase.presenter import Presenter

from turberfield.catchphrase.render import Action
from turberfield.catchphrase.render import Parameter
from turberfield.catchphrase.render import Renderer
from turberfield.catchphrase.render import Settings

from turberfield.dialogue.model import SceneScript
from turberfield.dialogue.model import Model

from turberfield.dialogue.types import DataObject
from turberfield.dialogue.types import EnumFactory
from turberfield.dialogue.types import Stateful

from turberfield.utils.assembly import Assembly
